@@ -2,7 +2,7 @@
  * _obnammodule.c -- Python extensions for Obna
  *
  * Copyright (C) 2008, 2009  Lars Wirzenius <liw@liw.fi>
- * 
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
@@ -21,7 +21,7 @@
 
 /*
  * This is a Python extension module written for Obnam, the backup
- * software. 
+ * software.
  *
  * This module provides a way to call the posix_fadvise function from
  * Python. Obnam uses this to use set the POSIX_FADV_SEQUENTIAL and
@@ -31,6 +31,8 @@
  */
 
 
+#define _FILE_OFFSET_BITS 64
+
 #include <Python.h>
 
 
@@ -43,10 +45,17 @@
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <sys/xattr.h>
 #include <unistd.h>
 #include <stdlib.h>
 
+#ifdef __FreeBSD__
+    #include <sys/extattr.h>
+    #define NO_NANOSECONDS 1
+#else
+    #include <sys/xattr.h>
+    #define NO_NANOSECONDS 0
+#endif
+
 
 static PyObject *
 fadvise_dontneed(PyObject *self, PyObject *args)
@@ -76,27 +85,54 @@ utimensat_wrapper(PyObject *self, PyObject *args)
     const char *filename;
     long atime_sec, atime_nsec;
     long mtime_sec, mtime_nsec;
+#if NO_NANOSECONDS
+    struct timeval tv[2];
+#else
     struct timespec tv[2];
-
-    if (!PyArg_ParseTuple(args, "sllll", 
-                          &filename, 
+#endif
+    if (!PyArg_ParseTuple(args, "sllll",
+                          &filename,
                           &atime_sec,
                           &atime_nsec,
                           &mtime_sec,
                           &mtime_nsec))
         return NULL;
 
+#if NO_NANOSECONDS
+    tv[0].tv_sec = atime_sec;
+    tv[0].tv_usec = atime_nsec / 1000;
+    tv[1].tv_sec = mtime_sec;
+    tv[1].tv_usec = mtime_nsec / 1000;
+    ret = lutimes(filename, tv);
+#else
     tv[0].tv_sec = atime_sec;
     tv[0].tv_nsec = atime_nsec;
     tv[1].tv_sec = mtime_sec;
     tv[1].tv_nsec = mtime_nsec;
     ret = utimensat(AT_FDCWD, filename, tv, AT_SYMLINK_NOFOLLOW);
+#endif
     if (ret == -1)
         ret = errno;
     return Py_BuildValue("i", ret);
 }
 
 
+/*
+ * Since we can't set nanosecond mtime and atimes on some platforms, also
+ * don't retrieve that level of precision from lstat(), so comparisons
+ * work.
+ */
+static unsigned long long
+remove_precision(unsigned long long nanoseconds)
+{
+#if NO_NANOSECONDS
+    return nanoseconds - (nanoseconds % 1000);
+#else
+    return nanoseconds;
+#endif
+}
+
+
 static PyObject *
 lstat_wrapper(PyObject *self, PyObject *args)
 {
@@ -110,24 +146,25 @@ lstat_wrapper(PyObject *self, PyObject *args)
     ret = lstat(filename, &st);
     if (ret == -1)
         ret = errno;
-    return Py_BuildValue("iLLLLLLLLLLLLLLLL", 
+
+    return Py_BuildValue("iKKKKKKKLLLLKLKLK",
                          ret,
-                         (long long) st.st_dev,
-                         (long long) st.st_ino,
-                         (long long) st.st_mode,
-                         (long long) st.st_nlink,
-                         (long long) st.st_uid,
-                         (long long) st.st_gid,
-                         (long long) st.st_rdev,
+                         (unsigned long long) st.st_dev,
+                         (unsigned long long) st.st_ino,
+                         (unsigned long long) st.st_mode,
+                         (unsigned long long) st.st_nlink,
+                         (unsigned long long) st.st_uid,
+                         (unsigned long long) st.st_gid,
+                         (unsigned long long) st.st_rdev,
                          (long long) st.st_size,
                          (long long) st.st_blksize,
                          (long long) st.st_blocks,
                          (long long) st.st_atim.tv_sec,
-                         (long long) st.st_atim.tv_nsec,
+                         remove_precision(st.st_atim.tv_nsec),
                          (long long) st.st_mtim.tv_sec,
-                         (long long) st.st_mtim.tv_nsec,
+                         remove_precision(st.st_mtim.tv_nsec),
                          (long long) st.st_ctim.tv_sec,
-                         (long long) st.st_ctim.tv_nsec);
+                         remove_precision(st.st_ctim.tv_nsec));
 }
 
 
@@ -137,16 +174,38 @@ llistxattr_wrapper(PyObject *self, PyObject *args)
     const char *filename;
     size_t bufsize;
     PyObject *o;
+    char* buf;
+    ssize_t n;
 
     if (!PyArg_ParseTuple(args, "s", &filename))
         return NULL;
 
+#ifdef __FreeBSD__
+    bufsize = extattr_list_link(filename, EXTATTR_NAMESPACE_USER, NULL, 0);
+    buf = malloc(bufsize);
+    n = extattr_list_link(filename, EXTATTR_NAMESPACE_USER, buf, bufsize);
+    if (n >= 0) {
+         /* Convert from length-prefixed BSD style to '\0'-suffixed
+            Linux style. */
+         size_t i = 0;
+         while (i < n) {
+             unsigned char length = (unsigned char) buf[i];
+	     memmove(buf + i, buf + i + 1, length);
+	     buf[i + length] = '\0';
+	     i += length + 1;
+	 }
+	 o = Py_BuildValue("s#", buf, (int) n);
+    } else {
+         o = Py_BuildValue("i", errno);
+    }
+    free(buf);
+#else
     bufsize = 0;
     o = NULL;
     do {
         bufsize += 1024;
-        char *buf = malloc(bufsize);
-        ssize_t n = llistxattr(filename, buf, bufsize);
+        buf = malloc(bufsize);
+        n = llistxattr(filename, buf, bufsize);
 
         if (n >= 0)
             o = Py_BuildValue("s#", buf, (int) n);
@@ -154,7 +213,7 @@ llistxattr_wrapper(PyObject *self, PyObject *args)
             o = Py_BuildValue("i", errno);
         free(buf);
     } while (o == NULL);
-    
+#endif
     return o;
 }
 
@@ -175,15 +234,18 @@ lgetxattr_wrapper(PyObject *self, PyObject *args)
     do {
         bufsize += 1024;
         char *buf = malloc(bufsize);
+#ifdef __FreeBSD__
+	int n = extattr_get_link(filename, EXTATTR_NAMESPACE_USER, attrname, buf, bufsize);
+#else
         ssize_t n = lgetxattr(filename, attrname, buf, bufsize);
-
-        if (n > 0)
+#endif
+        if (n >= 0)
             o = Py_BuildValue("s#", buf, (int) n);
         else if (n == -1 && errno != ERANGE)
             o = Py_BuildValue("i", errno);
         free(buf);
     } while (o == NULL);
-    
+
     return o;
 }
 
@@ -200,7 +262,11 @@ lsetxattr_wrapper(PyObject *self, PyObject *args)
     if (!PyArg_ParseTuple(args, "sss#", &filename, &name, &value, &size))
         return NULL;
 
+#ifdef __FreeBSD__
+    ret = extattr_set_link(filename, EXTATTR_NAMESPACE_USER, name, value, size);
+#else
     ret = lsetxattr(filename, name, value, size, 0);
+#endif
     if (ret == -1)
         ret = errno;
     return Py_BuildValue("i", ret);
@@ -208,7 +274,7 @@ lsetxattr_wrapper(PyObject *self, PyObject *args)
 
 
 static PyMethodDef methods[] = {
-    {"fadvise_dontneed",  fadvise_dontneed, METH_VARARGS, 
+    {"fadvise_dontneed",  fadvise_dontneed, METH_VARARGS,
      "Call posix_fadvise(2) with POSIX_FADV_DONTNEED argument."},
     {"utimensat", utimensat_wrapper, METH_VARARGS,
      "utimensat(2) wrapper."},
