@@ -31,18 +31,21 @@ import obnamlib
 class App(cliapp.Application):
 
     '''Main program for backup program.'''
-    
+
     def add_settings(self):
         devel_group = obnamlib.option_group['devel']
         perf_group = obnamlib.option_group['perf']
-    
+
         self.settings.string(['repository', 'r'], 'name of backup repository')
 
         self.settings.string(['client-name'], 'name of client (%default)',
                            default=self.deduce_client_name())
 
         self.settings.bytesize(['node-size'],
-                             'size of B-tree nodes on disk '
+                             'size of B-tree nodes on disk; only affects new '
+                                'B-trees so you may need to delete a client '
+                                'or repository to change this for existing '
+                                'repositories '
                                  '(default: %default)',
                               default=obnamlib.DEFAULT_NODE_SIZE,
                               group=perf_group)
@@ -69,7 +72,7 @@ class App(cliapp.Application):
                                 'add to filename patters for which trace '
                                 'debugging logging happens')
 
-        
+
         self.settings.integer(['idpath-depth'],
                               'depth of chunk id mapping',
                               default=obnamlib.IDPATH_DEPTH,
@@ -84,12 +87,13 @@ class App(cliapp.Application):
                               group=perf_group)
 
         self.settings.boolean(['quiet'], 'be silent')
+        self.settings.boolean(['verbose'], 'be verbose')
 
         self.settings.boolean(['pretend', 'dry-run', 'no-act'],
                            'do not actually change anything (works with '
                            'backup, forget and restore only, and may only '
                            'simulate approximately real behavior)')
-                           
+
         self.settings.string(['pretend-time'],
                              'pretend it is TIMESTAMP (YYYY-MM-DD HH:MM:SS); '
                                 'this is only useful for testing purposes',
@@ -123,7 +127,7 @@ class App(cliapp.Application):
         self.pm = obnamlib.PluginManager()
         self.pm.locations = [self.plugins_dir()]
         self.pm.plugin_arguments = (self,)
-        
+
         self.setup_hooks()
 
         self.fsf = obnamlib.VfsFactory()
@@ -132,6 +136,8 @@ class App(cliapp.Application):
         self.pm.enable_plugins()
         self.hooks.call('plugins-loaded')
 
+        self.settings['log-level'] = 'info'
+
     def deduce_client_name(self):
         return socket.gethostname()
 
@@ -163,26 +169,18 @@ class App(cliapp.Application):
         try:
             if self.settings['quiet']:
                 self.ts.disable()
-            self.log_config()
             for pattern in self.settings['trace']:
                 tracing.trace_add_pattern(pattern)
             self.hooks.call('config-loaded')
             cliapp.Application.process_args(self, args)
             self.hooks.call('shutdown')
-            logging.info('Obnam ends')
         except larch.Error, e:
             logging.critical(str(e))
             sys.stderr.write('ERROR: %s\n' % str(e))
             sys.exit(1)
 
-    def log_config(self):
-        '''Log current configuration into the log file.'''
-        f = StringIO.StringIO()
-        self.settings.dump_config(f)
-        logging.debug('Current configuration:\n%s' % f.getvalue())
-
     def setup_ttystatus(self):
-        self.ts = ttystatus.TerminalStatus(period=0.25)
+        self.ts = ttystatus.TerminalStatus(period=0.1)
         if self.settings['quiet']:
             self.ts.disable()
 
@@ -197,7 +195,7 @@ class App(cliapp.Application):
             repofs.connect()
         else:
             repofs.reinit(repopath)
-        return obnamlib.Repository(repofs, 
+        return obnamlib.Repository(repofs,
                                     self.settings['node-size'],
                                     self.settings['upload-queue-size'],
                                     self.settings['lru-size'],
@@ -211,10 +209,10 @@ class App(cliapp.Application):
 
     def time(self):
         '''Return current time in seconds since epoch.
-        
+
         This is a wrapper around time.time() so that it can be overridden
         with the --pretend-time setting.
-        
+
         '''
 
         s = self.settings['pretend-time']
