@@ -14,22 +14,25 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 
+import errno
 import grp
+import logging
 import os
 import pwd
 import stat
 import struct
+import tracing
 
 import obnamlib
 
 
 metadata_verify_fields = (
-    'st_mode', 'st_mtime_sec', 'st_mtime_nsec', 
+    'st_mode', 'st_mtime_sec', 'st_mtime_nsec',
     'st_nlink', 'st_size', 'st_uid', 'groupname', 'username', 'target',
     'xattr',
 )
 metadata_fields = metadata_verify_fields + (
-    'st_blocks', 'st_dev', 'st_gid', 'st_ino',  'st_atime_sec', 
+    'st_blocks', 'st_dev', 'st_gid', 'st_ino',  'st_atime_sec',
     'st_atime_nsec', 'md5',
 )
 
@@ -37,16 +40,16 @@ metadata_fields = metadata_verify_fields + (
 class Metadata(object):
 
     '''Represent metadata for a filesystem entry.
-    
+
     The metadata for a filesystem entry (file, directory, device, ...)
     consists of its stat(2) result, plus ACL and xattr.
-    
+
     This class represents them as fields.
-    
+
     We do not store all stat(2) fields. Here's a commentary on all fields:
-    
+
         field?          stored? why
-    
+
         st_atime_sec    yes     mutt compares atime, mtime to see ifmsg is new
         st_atime_nsec   yes     mutt compares atime, mtime to see ifmsg is new
         st_blksize      no      no way to restore, not useful backed up
@@ -64,18 +67,18 @@ class Metadata(object):
         st_uid          yes     used to restored ownership
 
     The field 'target' stores the target of a symlink.
-        
+
     Additionally, the fields 'groupname' and 'username' are stored. They
     contain the textual names that correspond to st_gid and st_uid. When
     restoring, the names will be preferred by default.
-    
+
     The 'md5' field optionally stores the whole-file checksum for the file.
-    
+
     The 'xattr' field optionally stores extended attributes encoded as
     a binary blob.
-    
+
     '''
-    
+
     def __init__(self, **kwargs):
         for field in metadata_fields:
             setattr(self, field, None)
@@ -118,7 +121,7 @@ def _cached_getpwuid(uid): # pragma: no cover
     if uid not in _uid_to_username:
         _uid_to_username[uid] = pwd.getpwuid(uid)
     return _uid_to_username[uid]
-    
+
 _gid_to_groupname = {}
 def _cached_getgrgid(gid): # pragma: no cover
     if gid not in _gid_to_groupname:
@@ -127,10 +130,41 @@ def _cached_getgrgid(gid): # pragma: no cover
 
 
 def get_xattrs_as_blob(fs, filename): # pragma: no cover
-    names = fs.llistxattr(filename)
+    tracing.trace('filename=%s' % filename)
+
+    try:
+        names = fs.llistxattr(filename)
+    except (OSError, IOError), e:
+        if e.errno in (errno.EOPNOTSUPP, errno.EACCES):
+            return None
+        raise
+    tracing.trace('names=%s' % repr(names))
     if not names:
         return None
-    values = [fs.lgetxattr(filename, name) for name in names]
+
+    values = []
+    for name in names[:]:
+        tracing.trace('trying name %s' % repr(name))
+        try:
+            value = fs.lgetxattr(filename, name)
+        except OSError, e:
+            # On btrfs, at least, this can happen: the filesystem returns
+            # a list of attribute names, but then fails when looking up
+            # the value for one or more of the names. We pretend that the
+            # name was never returned in that case.
+            #
+            # Obviously this can happen due to race conditions as well.
+            if e.errno == errno.ENODATA:
+                names.remove(name)
+                logging.warning(
+                    '%s has extended attribute named %s without value, '
+                    'ignoring attribute' % (filename, name))
+            else:
+                raise
+        else:
+            tracing.trace('lgetxattr(%s)=%s' % (name, value))
+            values.append(value)
+    assert len(names) == len(values)
 
     name_blob = ''.join('%s\0' % name for name in names)
 
@@ -138,7 +172,7 @@ def get_xattrs_as_blob(fs, filename): # pragma: no cover
     fmt = '!' + 'Q' * len(values)
     value_blob = struct.pack(fmt, *lengths) + ''.join(values)
 
-    return ('%s%s%s' % 
+    return ('%s%s%s' %
             (struct.pack('!Q', len(name_blob)),
              name_blob,
              value_blob))
@@ -154,7 +188,7 @@ def set_xattrs_from_blob(fs, filename, blob): # pragma: no cover
     fmt = '!' + 'Q' * len(names)
     lengths_size = sizesize * len(names)
     lengths = struct.unpack(fmt, value_blob[:lengths_size])
-    
+
     pos = lengths_size
     for i, name in enumerate(names):
         value = value_blob[pos:pos + lengths[i]]
@@ -201,25 +235,35 @@ def set_metadata(fs, filename, metadata, getuid=None):
     fields: we assume the caller will change st_uid, st_gid accordingly
     if they want to mess with things. This makes the user take care
     of error situations and looking up user preferences.
-    
+
     '''
 
-    if stat.S_ISLNK(metadata.st_mode):
+    symlink = stat.S_ISLNK(metadata.st_mode)
+    if symlink:
         fs.symlink(metadata.target, filename)
+
+    # Set owner before mode, so that a setuid bit does not get reset.
+    getuid = getuid or os.getuid
+    if getuid() == 0:
+        fs.lchown(filename, metadata.st_uid, metadata.st_gid)
+
+    # If we are not the owner, and not root, do not restore setuid/setgid.
+    mode = metadata.st_mode
+    if getuid() not in (0, metadata.st_uid): # pragma: no cover
+        mode = mode & (~stat.S_ISUID)
+        mode = mode & (~stat.S_ISGID)
+    if symlink:
+        fs.chmod_symlink(filename, mode)
     else:
-        fs.chmod(filename, metadata.st_mode)
+        fs.chmod_not_symlink(filename, mode)
 
     if metadata.xattr: # pragma: no cover
         set_xattrs_from_blob(fs, filename, metadata.xattr)
 
-    fs.lutimes(filename, metadata.st_atime_sec, metadata.st_atime_nsec, 
+    fs.lutimes(filename, metadata.st_atime_sec, metadata.st_atime_nsec,
                metadata.st_mtime_sec, metadata.st_mtime_nsec)
 
-    getuid = getuid or os.getuid
-    if getuid() == 0:
-        fs.lchown(filename, metadata.st_uid, metadata.st_gid)
-    
-    
+
 metadata_format = struct.Struct('!Q' +  # flags
                                 'Q' +   # st_mode
                                 'qQ' +  # st_mtime_sec and _nsec
@@ -244,25 +288,46 @@ def encode_metadata(metadata):
         if getattr(metadata, name) is not None:
             flags |= (1 << i)
 
-    packed = metadata_format.pack(flags,
-                                  metadata.st_mode or 0,
-                                  metadata.st_mtime_sec or 0,
-                                  metadata.st_mtime_nsec or 0,
-                                  metadata.st_atime_sec or 0,
-                                  metadata.st_atime_nsec or 0,
-                                  metadata.st_nlink or 0,
-                                  metadata.st_size or 0,
-                                  metadata.st_uid or 0,
-                                  metadata.st_gid or 0,
-                                  metadata.st_dev or 0,
-                                  metadata.st_ino or 0,
-                                  metadata.st_blocks or 0,
-                                  len(metadata.groupname or ''),
-                                  len(metadata.username or ''),
-                                  len(metadata.target or ''),
-                                  len(metadata.md5 or ''),
-                                  len(metadata.xattr or ''))
-    return (packed + 
+    try:
+        packed = metadata_format.pack(flags,
+                                      metadata.st_mode or 0,
+                                      metadata.st_mtime_sec or 0,
+                                      metadata.st_mtime_nsec or 0,
+                                      metadata.st_atime_sec or 0,
+                                      metadata.st_atime_nsec or 0,
+                                      metadata.st_nlink or 0,
+                                      metadata.st_size or 0,
+                                      metadata.st_uid or 0,
+                                      metadata.st_gid or 0,
+                                      metadata.st_dev or 0,
+                                      metadata.st_ino or 0,
+                                      metadata.st_blocks or 0,
+                                      len(metadata.groupname or ''),
+                                      len(metadata.username or ''),
+                                      len(metadata.target or ''),
+                                      len(metadata.md5 or ''),
+                                      len(metadata.xattr or ''))
+    except TypeError, e: # pragma: no cover
+        logging.error('ERROR: Packing error due to %s' % str(e))
+        logging.error('ERROR: st_mode=%s' % repr(metadata.st_mode))
+        logging.error('ERROR: st_mtime_sec=%s' % repr(metadata.st_mtime_sec))
+        logging.error('ERROR: st_mtime_nsec=%s' % repr(metadata.st_mtime_nsec))
+        logging.error('ERROR: st_atime_sec=%s' % repr(metadata.st_atime_sec))
+        logging.error('ERROR: st_atime_nsec=%s' % repr(metadata.st_atime_nsec))
+        logging.error('ERROR: st_nlink=%s' % repr(metadata.st_nlink))
+        logging.error('ERROR: st_size=%s' % repr(metadata.st_size))
+        logging.error('ERROR: st_uid=%s' % repr(metadata.st_uid))
+        logging.error('ERROR: st_gid=%s' % repr(metadata.st_gid))
+        logging.error('ERROR: st_dev=%s' % repr(metadata.st_dev))
+        logging.error('ERROR: st_ino=%s' % repr(metadata.st_ino))
+        logging.error('ERROR: st_blocks=%s' % repr(metadata.st_blocks))
+        logging.error('ERROR: groupname=%s' % repr(metadata.groupname))
+        logging.error('ERROR: username=%s' % repr(metadata.username))
+        logging.error('ERROR: target=%s' % repr(metadata.target))
+        logging.error('ERROR: md5=%s' % repr(metadata.md5))
+        logging.error('ERROR: xattr=%s' % repr(metadata.xattr))
+        raise
+    return (packed +
              (metadata.groupname or '') +
              (metadata.username or '') +
              (metadata.target or '') +
@@ -276,7 +341,7 @@ def decode_metadata(encoded):
     flags = items[0]
     pos = [1, metadata_format.size]
     metadata = obnamlib.Metadata()
-    
+
     def is_present(field):
         i = obnamlib.metadata_fields.index(field)
         return (flags & (1 << i)) != 0
@@ -294,7 +359,7 @@ def decode_metadata(encoded):
 
     def decode_string(field):
         decode(field, 1, True, lambda i, o: encoded[o:o + items[i]])
-    
+
     decode_integer('st_mode')
     decode_integer('st_mtime_sec')
     decode_integer('st_mtime_nsec')
@@ -312,6 +377,6 @@ def decode_metadata(encoded):
     decode_string('target')
     decode_string('md5')
     decode_string('xattr')
-    
+
     return metadata
 
