@@ -17,6 +17,7 @@
 import logging
 import os
 import stat
+import time
 import ttystatus
 
 import obnamlib
@@ -25,23 +26,23 @@ import obnamlib
 class Hardlinks(object):
 
     '''Keep track of inodes with unrestored hardlinks.'''
-    
+
     def __init__(self):
         self.inodes = dict()
-        
+
     def key(self, metadata):
         return '%s:%s' % (metadata.st_dev, metadata.st_ino)
-        
+
     def add(self, filename, metadata):
         self.inodes[self.key(metadata)] = (filename, metadata.st_nlink)
-        
+
     def filename(self, metadata):
         key = self.key(metadata)
         if key in self.inodes:
             return self.inodes[key][0]
         else:
             return None
-        
+
     def forget(self, metadata):
         key = self.key(metadata)
         filename, nlinks = self.inodes[key]
@@ -61,12 +62,12 @@ class RestorePlugin(obnamlib.ObnamPlugin):
     # just b if b is an absolute path.
 
     def enable(self):
-        self.app.add_subcommand('restore', self.restore, 
-                                arg_synopsis='[FILE]...')
+        self.app.add_subcommand('restore', self.restore,
+                                arg_synopsis='[DIRECTORY]...')
         self.app.settings.string(['to'], 'where to restore')
-        self.app.settings.string(['generation'], 
+        self.app.settings.string_list(['generation'],
                                 'which generation to restore',
-                                 default='latest')
+                                 default=['latest'])
 
     @property
     def write_ok(self):
@@ -77,7 +78,7 @@ class RestorePlugin(obnamlib.ObnamPlugin):
         self.app.ts['total'] = 0
         self.app.ts['current-bytes'] = 0
         self.app.ts['total-bytes'] = 0
-        
+
         self.app.ts.format('%RemainingTime(current-bytes,total-bytes) '
                            '%Counter(current) files '
                            '%ByteSize(current-bytes) '
@@ -92,15 +93,19 @@ class RestorePlugin(obnamlib.ObnamPlugin):
         self.app.settings.require('generation')
         self.app.settings.require('to')
 
-        logging.debug('restoring generation %s' % 
+        logging.debug('restoring generation %s' %
                         self.app.settings['generation'])
         logging.debug('restoring to %s' % self.app.settings['to'])
-    
+
         logging.debug('restoring what: %s' % repr(args))
         if not args:
             logging.debug('no args given, so restoring everything')
             args = ['/']
-    
+
+        self.downloaded_bytes = 0
+        self.file_count = 0
+        self.started = time.time()
+
         self.repo = self.app.open_repository()
         self.repo.open_client(self.app.settings['client-name'])
         if self.write_ok:
@@ -110,10 +115,15 @@ class RestorePlugin(obnamlib.ObnamPlugin):
             self.fs = None # this will trigger error if we try to really write
 
         self.hardlinks = Hardlinks()
-        
+
         self.errors = False
-        
-        gen = self.repo.genspec(self.app.settings['generation'])
+
+        generations = self.app.settings['generation']
+        if len(generations) != 1:
+            raise obnamlib.Error(
+                'The restore command wants exactly one generation option')
+
+        gen = self.repo.genspec(generations[0])
 
         self.configure_ttystatus()
         self.app.ts['total'] = self.repo.client.get_generation_file_count(gen)
@@ -128,19 +138,27 @@ class RestorePlugin(obnamlib.ObnamPlugin):
         self.repo.fs.close()
         if self.write_ok:
             self.fs.close()
-        
+
+        self.app.ts.clear()
+        self.report_stats()
+
         self.app.ts.finish()
-                
+
         if self.errors:
             raise obnamlib.Error('There were errors when restoring')
 
     def restore_something(self, gen, root):
         for pathname, metadata in self.repo.walk(gen, root, depth_first=True):
+            self.file_count += 1
             self.app.ts['current'] = pathname
+            self.restore_safely(gen, pathname, metadata)
+
+    def restore_safely(self, gen, pathname, metadata):
+        try:
             dirname = os.path.dirname(pathname)
             if self.write_ok and not self.fs.exists('./' + dirname):
                 self.fs.makedirs('./' + dirname)
-    
+
             set_metadata = True
             if metadata.isdir():
                 self.restore_dir(gen, pathname, metadata)
@@ -157,7 +175,20 @@ class RestorePlugin(obnamlib.ObnamPlugin):
             else:
                 self.restore_first_link(gen, pathname, metadata)
             if set_metadata and self.write_ok:
-                obnamlib.set_metadata(self.fs, './' + pathname, metadata)
+                try:
+                    obnamlib.set_metadata(self.fs, './' + pathname, metadata)
+                except (IOError, OSError), e:
+                    msg = ('Could not set metadata: %s: %d: %s' %
+                            (pathname, e.errno, e.strerror))
+                    logging.error(msg)
+                    self.app.ts.notify(msg)
+                    self.errors = True
+        except Exception, e:
+            # Reaching this code path means we've hit a bug, so we log a full traceback.
+            msg = "Failed to restore %s:" % (pathname,)
+            logging.exception(msg)
+            self.app.ts.notify(msg + " " + str(e))
+            self.errors = True
 
     def restore_dir(self, gen, root, metadata):
         logging.debug('restoring dir %s' % root)
@@ -171,7 +202,7 @@ class RestorePlugin(obnamlib.ObnamPlugin):
         if self.write_ok:
             self.fs.link('./' + link, './' + filename)
             self.hardlinks.forget(metadata)
-        
+
     def restore_symlink(self, gen, filename, metadata):
         logging.debug('restoring symlink %s' % filename)
 
@@ -182,12 +213,14 @@ class RestorePlugin(obnamlib.ObnamPlugin):
             self.restore_fifo(gen, filename, metadata)
         elif stat.S_ISSOCK(metadata.st_mode):
             self.restore_socket(gen, filename, metadata)
+        elif stat.S_ISBLK(metadata.st_mode) or stat.S_ISCHR(metadata.st_mode):
+            self.restore_device(gen, filename, metadata)
         else:
-            msg = ('Unknown file type: %s (%o)' % 
+            msg = ('Unknown file type: %s (%o)' %
                    (filename, metadata.st_mode))
             logging.error(msg)
             self.app.ts.notify(msg)
-        
+
     def restore_regular_file(self, gen, filename, metadata):
         logging.debug('restoring regular %s' % filename)
         if self.write_ok:
@@ -202,6 +235,7 @@ class RestorePlugin(obnamlib.ObnamPlugin):
                 else:
                     f.write(contents)
                     summer.update(contents)
+                    self.downloaded_bytes += len(contents)
             except obnamlib.MissingFilterError, e:
                 msg = 'Missing filter error during restore: %s' % filename
                 logging.error(msg)
@@ -223,6 +257,7 @@ class RestorePlugin(obnamlib.ObnamPlugin):
             data = self.repo.get_chunk(chunkid)
             self.verify_chunk_checksum(data, chunkid)
             checksummer.update(data)
+            self.downloaded_bytes += len(data)
             if len(data) != len(zeroes):
                 zeroes = '\0' * len(data)
             if data == zeroes:
@@ -259,3 +294,66 @@ class RestorePlugin(obnamlib.ObnamPlugin):
         if self.write_ok:
             self.fs.mknod('./' + filename, metadata.st_mode)
 
+    def restore_device(self, gen, filename, metadata):
+        logging.debug('restoring device %s' % filename)
+        if self.write_ok:
+            self.fs.mknod('./' + filename, metadata.st_mode)
+
+    def report_stats(self):
+        size_table = [
+            (1024**4, 'TiB'),
+            (1024**3, 'GiB'),
+            (1024**2, 'MiB'),
+            (1024**1, 'KiB'),
+            (0, 'B')
+        ]
+
+        for size_base, size_unit in size_table:
+            if self.downloaded_bytes >= size_base:
+                if size_base > 0:
+                    size_amount = (float(self.downloaded_bytes) /
+                                    float(size_base))
+                else:
+                    size_amount = float(self.downloaded_bytes)
+                break
+
+        speed_table = [
+            (1024**3, 'GiB/s'),
+            (1024**2, 'MiB/s'),
+            (1024**1, 'KiB/s'),
+            (0, 'B/s')
+        ]
+        duration = time.time() - self.started
+        speed = float(self.downloaded_bytes) / duration
+        for speed_base, speed_unit in speed_table:
+            if speed >= speed_base:
+                if speed_base > 0:
+                    speed_amount = speed / speed_base
+                else:
+                    speed_amount = speed
+                break
+
+        duration_string = ''
+        seconds = duration
+        if seconds >= 3600:
+            duration_string += '%dh' % int(seconds/3600)
+            seconds %= 3600
+        if seconds >= 60:
+            duration_string += '%dm' % int(seconds/60)
+            seconds %= 60
+        if seconds > 0:
+            duration_string += '%ds' % round(seconds)
+
+        logging.info('Restore performance statistics:')
+        logging.info('* files restored: %s' % self.file_count)
+        logging.info('* downloaded data: %s bytes (%s %s)' %
+                        (self.downloaded_bytes, size_amount, size_unit))
+        logging.info('* duration: %s s' % duration)
+        logging.info('* average speed: %s %s' % (speed_amount, speed_unit))
+        self.app.ts.notify(
+            'Restored %d files, '
+            'downloaded %.1f %s in %s at %.1f %s average speed' %
+                (self.file_count,
+                 size_amount, size_unit,
+                 duration_string, speed_amount, speed_unit))
+
