@@ -25,6 +25,7 @@ import socket
 import stat
 import subprocess
 import time
+import traceback
 import urlparse
 
 # As of 2010-07-10, Debian's paramiko package triggers
@@ -40,30 +41,27 @@ with warnings.catch_warnings():
 import obnamlib
 
 
-DEFAULT_SSH_PORT = 22
-
-
 def ioerror_to_oserror(method):
     '''Decorator to convert an IOError exception to OSError.
-    
+
     Python's os.* raise OSError, mostly, but paramiko's corresponding
     methods raise IOError. This decorator fixes that.
-    
+
     '''
-    
+
     def helper(self, filename, *args, **kwargs):
         try:
             return method(self, filename, *args, **kwargs)
         except IOError, e:
             raise OSError(e.errno, e.strerror or str(e), filename)
-    
+
     return helper
 
 
 class SSHChannelAdapter(object):
 
     '''Take an ssh subprocess and pretend it is a paramiko Channel.'''
-    
+
     # This is inspired by the ssh.py module in bzrlib.
 
     def __init__(self, proc):
@@ -88,7 +86,7 @@ class SSHChannelAdapter(object):
 
     def close(self):
         logging.debug('SSHChannelAdapter.close called')
-        for func in [self.proc.stdin.close, self.proc.stdout.close, 
+        for func in [self.proc.stdin.close, self.proc.stdout.close,
                      self.proc.wait]:
             try:
                 func()
@@ -99,11 +97,11 @@ class SSHChannelAdapter(object):
 class SftpFS(obnamlib.VirtualFileSystem):
 
     '''A VFS implementation for SFTP.
-    
-    
-    
+
+
+
     '''
-    
+
     # 32 KiB is the chunk size that gives me the fastest speed
     # for sftp transfers. I don't know why the size matters.
     chunk_size = 32 * 1024
@@ -113,7 +111,11 @@ class SftpFS(obnamlib.VirtualFileSystem):
         self.sftp = None
         self.settings = settings
         self._roundtrips = 0
+        self._initial_dir = None
         self.reinit(baseurl, create=create)
+        # Backwards compatibility with old, deprecated option:
+        if settings and settings['strict-ssh-host-keys']:
+            settings["ssh-host-keys-check"] = "yes"
 
     def _delay(self):
         self._roundtrips += 1
@@ -126,7 +128,7 @@ class SftpFS(obnamlib.VirtualFileSystem):
         obnamlib.VirtualFileSystem.log_stats(self)
         logging.info('VFS: baseurl=%s roundtrips=%s' %
                          (self.baseurl, self._roundtrips))
-        
+
     def _to_string(self, str_or_unicode):
         if type(str_or_unicode) is unicode:
             return str_or_unicode.encode('utf-8')
@@ -141,32 +143,43 @@ class SftpFS(obnamlib.VirtualFileSystem):
             # for the best
             pass
         self.create_path_if_missing = False # only create once
-        
+
     def connect(self):
         try_openssh = not self.settings or not self.settings['pure-paramiko']
         if not try_openssh or not self._connect_openssh():
             self._connect_paramiko()
         if self.create_path_if_missing:
             self._create_root_if_missing()
+        self.chdir('.')
+        self._initial_dir = self.getcwd()
         self.chdir(self.path)
 
     def _connect_openssh(self):
-        args = ['ssh',
-                '-oForwardX11=no', '-oForwardAgent=no',
+        executable = 'ssh'
+        args = ['-oForwardX11=no', '-oForwardAgent=no',
                 '-oClearAllForwardings=yes', '-oProtocol=2',
-                '-p', str(self.port),
-                '-l', self.user,
                 '-s']
+        if self.settings and self.settings['ssh-command']:
+            executable = self.settings["ssh-command"]
+        # default user/port from ssh (could be a per host configuration)
+        if self.port:
+            args += ['-p', str(self.port)]
+        if self.user:
+            args += ['-l', self.user]
         if self.settings and self.settings['ssh-key']:
             args += ['-i', self.settings['ssh-key']]
-        if self.settings and self.settings['strict-ssh-host-keys']:
-            args += ['-o', 'StrictHostKeyChecking=yes']
+        if (self.settings and
+            self.settings['ssh-host-keys-check'] != "ssh-config"):
+            value = self.settings['ssh-host-keys-check']
+            args += ['-o', 'StrictHostKeyChecking=%s' % (value,)]
         if self.settings and self.settings['ssh-known-hosts']:
-            args += ['-o', 
-                     'UserKnownHostsFile=%s' % 
+            args += ['-o',
+                     'UserKnownHostsFile=%s' %
                         self.settings['ssh-known-hosts']]
         args += [self.host, 'sftp']
 
+        # prepend the executable to the argument list
+        args.insert(0, executable)
         logging.debug('executing openssh: %s' % args)
         try:
             proc = subprocess.Popen(args,
@@ -181,8 +194,13 @@ class SftpFS(obnamlib.VirtualFileSystem):
         return True
 
     def _connect_paramiko(self):
-        logging.debug('connect_paramiko: host=%s port=%s' % (self.host, self.port))
-        self.transport = paramiko.Transport((self.host, self.port))
+        logging.debug(
+            'connect_paramiko: host=%s port=%s' % (self.host, self.port))
+        if self.port:
+            remote = (self.host, self.port)
+        else:
+            remote = (self.host)
+        self.transport = paramiko.Transport(remote)
         self.transport.connect()
         logging.debug('connect_paramiko: connected')
         try:
@@ -207,7 +225,7 @@ class SftpFS(obnamlib.VirtualFileSystem):
 
         known_keys = known_hosts.lookup(hostname)
         if known_keys is None:
-            if self.settings['strict-ssh-host-keys']:
+            if self.settings['ssh-host-keys-check'] == 'yes':
                 raise obnamlib.Error('No known host key for %s' % hostname)
             logging.warning('No known host keys for %s; accepting offered key'
                             % hostname)
@@ -215,20 +233,22 @@ class SftpFS(obnamlib.VirtualFileSystem):
 
         offered_type = offered_key.get_name()
         if not known_keys.has_key(offered_type):
-            if self.settings['strict-ssh-host-keys']:
-                raise obnamlib.Error('No known type %s host key for %s' % 
+            if self.settings['ssh-host-keys-check'] == 'yes':
+                raise obnamlib.Error('No known type %s host key for %s' %
                                      (offered_type, hostname))
             logging.warning('No known host key of type %s for %s; accepting '
                             'offered key' % (offered_type, hostname))
-        
+
         known_key = known_keys[offered_type]
         if offered_key != known_key:
             raise obnamlib.Error('SSH server %s offered wrong public key' %
                                  hostname)
-            
-        logging.debug('Host key for %s OK' % hostname)        
-    
+
+        logging.debug('Host key for %s OK' % hostname)
+
     def _authenticate(self, username):
+        if not username:
+            username = self._get_username()
         for key in self._find_auth_keys():
             try:
                 self.transport.auth_publickey(username, key)
@@ -276,14 +296,23 @@ class SftpFS(obnamlib.VirtualFileSystem):
         if '@' in netloc:
             user, netloc = netloc.split('@', 1)
         else:
-            user = self._get_username()
+            user = None
 
         if ':' in netloc:
             host, port = netloc.split(':', 1)
-            port = int(port)
+            if port == '':
+                port = None
+            else:
+                try:
+                    port = int(port)
+                except ValueError, e:
+                    msg = ('Invalid port number %s in %s: %s' %
+                            (port, baseurl, str(e)))
+                    logging.error(msg)
+                    raise obnamlib.Error(msg)
         else:
             host = netloc
-            port = DEFAULT_SSH_PORT
+            port = None
 
         if path.startswith('/~/'):
             path = path[3:]
@@ -295,10 +324,12 @@ class SftpFS(obnamlib.VirtualFileSystem):
         self.create_path_if_missing = create
 
         self._delay()
-        
+
         if self.sftp:
             if create:
                 self._create_root_if_missing()
+            logging.debug('chdir to %s' % path)
+            self.sftp.chdir(self._initial_dir)
             self.sftp.chdir(path)
 
     def _get_username(self):
@@ -350,17 +381,17 @@ class SftpFS(obnamlib.VirtualFileSystem):
         # fixed, this code can be removed.
         st.st_mtime_sec = self._force_32bit_timestamp(st.st_mtime)
         st.st_atime_sec = self._force_32bit_timestamp(st.st_atime)
-        
+
         # Within Obnam, we pretend stat results have st_Xtime_sec and
         # st_Xtime_nsec, but not st_Xtime. Remove those fields.
         del st.st_mtime
         del st.st_atime
-        
+
         # We only get integer timestamps, so set these explicitly to 0.
         st.st_mtime_nsec = 0
         st.st_atime_nsec = 0
 
-        return st        
+        return st
 
     @ioerror_to_oserror
     def listdir2(self, pathname):
@@ -397,7 +428,7 @@ class SftpFS(obnamlib.VirtualFileSystem):
             return stat.S_ISDIR(st.st_mode)
 
     def mknod(self, pathname, mode):
-        # SFTP does not provide an mknod, so we can't do this. We 
+        # SFTP does not provide an mknod, so we can't do this. We
         # raise an exception, so upper layers can handle this (we _could_
         # just fail silently, but that would be silly.)
         raise NotImplementedError('mknod on SFTP: %s' % pathname)
@@ -406,7 +437,7 @@ class SftpFS(obnamlib.VirtualFileSystem):
     def mkdir(self, pathname):
         self._delay()
         self.sftp.mkdir(pathname)
-        
+
     @ioerror_to_oserror
     def makedirs(self, pathname):
         parent = os.path.dirname(pathname)
@@ -418,7 +449,7 @@ class SftpFS(obnamlib.VirtualFileSystem):
     def rmdir(self, pathname):
         self._delay()
         self.sftp.rmdir(pathname)
-        
+
     @ioerror_to_oserror
     def remove(self, pathname):
         self._delay()
@@ -437,7 +468,7 @@ class SftpFS(obnamlib.VirtualFileSystem):
         self._delay()
         self._remove_if_exists(new)
         self.sftp.rename(old, new)
-    
+
     @ioerror_to_oserror
     def lstat(self, pathname):
         self._delay()
@@ -452,12 +483,19 @@ class SftpFS(obnamlib.VirtualFileSystem):
             logging.warning('NOT changing ownership of symlink %s' % pathname)
         else:
             self.sftp.chown(pathname, uid, gid)
-        
+
     @ioerror_to_oserror
-    def chmod(self, pathname, mode):
+    def chmod_symlink(self, pathname, mode):
+        # SFTP and/or paramiko don't have lchmod at all, so we can't
+        # actually do this. However, we at least check that pathname
+        # exists.
+        self.lstat(pathname)
+
+    @ioerror_to_oserror
+    def chmod_not_symlink(self, pathname, mode):
         self._delay()
         self.sftp.chmod(pathname, mode)
-        
+
     @ioerror_to_oserror
     def lutimes(self, pathname, atime_sec, atime_nsec, mtime_sec, mtime_nsec):
         # FIXME: This does not work for symlinks!
@@ -507,7 +545,13 @@ class SftpFS(obnamlib.VirtualFileSystem):
         try:
             f = self.open(pathname, 'wx')
         except (IOError, OSError), e:
-            if e.errno != errno.ENOENT:
+            # When the path to the file to be written does not
+            # exist, we try to create the directories below. Note that
+            # some SFTP servers return EACCES instead of ENOENT
+            # when the path to the file does not exist, so we
+            # do not raise an exception here for both ENOENT
+            # and EACCES.
+            if e.errno != errno.ENOENT and e.errno != errno.EACCES:
                 raise
             dirname = os.path.dirname(pathname)
             self.makedirs(dirname)
@@ -518,7 +562,7 @@ class SftpFS(obnamlib.VirtualFileSystem):
 
     def _tempfile(self, dirname):
         '''Create a new file with a random name, return file handle and name.'''
-        
+
         if dirname:
             try:
                 self.makedirs(dirname)
@@ -548,7 +592,7 @@ class SftpFS(obnamlib.VirtualFileSystem):
         self._write_helper(f, contents)
         f.close()
         self.rename(tempname, pathname)
-        
+
     def _write_helper(self, f, contents):
         for pos in range(0, len(contents), self.chunk_size):
             chunk = contents[pos:pos + self.chunk_size]
@@ -575,10 +619,21 @@ class SftpPlugin(obnamlib.ObnamPlugin):
                                  group=ssh_group)
 
         self.app.settings.boolean(['strict-ssh-host-keys'],
-                                  'require that the ssh host key must be '
-                                    'known and correct to be accepted; '
-                                    'default is to accept unknown keys',
-                                  group=ssh_group)
+                                  'DEPRECATED, use --ssh-host-keys-check '
+                                    'instead',
+                                 group=ssh_group)
+
+        self.app.settings.choice(['ssh-host-keys-check'],
+                                 ['ssh-config', 'yes', 'no', 'ask'],
+                                  'If "yes", require that the ssh host key must '
+                                    'be known and correct to be accepted. If '
+                                    '"no", do not require that. If "ask", the '
+                                    'user is interactively asked to accept new '
+                                    'hosts. The default ("ssh-config") is to '
+                                    'rely on the settings of the underlying '
+                                    'SSH client',
+                                 metavar='VALUE',
+                                 group=ssh_group)
 
         self.app.settings.string(['ssh-known-hosts'],
                                  'filename of the user\'s known hosts file '
@@ -588,6 +643,13 @@ class SftpPlugin(obnamlib.ObnamPlugin):
                                     os.path.expanduser('~/.ssh/known_hosts'),
                                  group=ssh_group)
 
+        self.app.settings.string(['ssh-command'],
+                                 'alternative executable to be used instead '
+                                    'of "ssh" (full path is allowed, no '
+                                    'arguments may be added)',
+                                 metavar='EXECUTABLE',
+                                 group=ssh_group)
+
         self.app.settings.boolean(['pure-paramiko'],
                                  'do not use openssh even if available, '
                                     'use paramiko only instead',
