@@ -23,11 +23,16 @@ import math
 import os
 import pwd
 import tempfile
+import time
 import tracing
 
 import obnamlib
 
 
+# O_NOATIME is Linux specific:
+EXTRA_OPEN_FLAGS = getattr(os, "O_NOATIME", 0)
+
+
 class LocalFSFile(file):
 
     def read(self, amount=-1):
@@ -48,20 +53,27 @@ class LocalFSFile(file):
 class LocalFS(obnamlib.VirtualFileSystem):
 
     """A VFS implementation for local filesystems."""
-    
+
     chunk_size = 1024 * 1024
-    
+
     def __init__(self, baseurl, create=False):
         tracing.trace('baseurl=%s', baseurl)
         tracing.trace('create=%s', create)
         obnamlib.VirtualFileSystem.__init__(self, baseurl)
         self.reinit(baseurl, create=create)
 
+        # For checking that we do not unlock something we didn't lock
+        # ourselves.
+        self.our_locks = set()
+
         # For testing purposes, allow setting a limit on write operations
         # after which an exception gets raised. If set to None, no crash.
         self.crash_limit = None
         self.crash_counter = 0
 
+        # Do we have lchmod?
+        self.got_lchmod = hasattr(os, 'lchmod')
+
     def maybe_crash(self): # pragma: no cover
         if self.crash_limit is not None:
             self.crash_counter += 1
@@ -70,7 +82,7 @@ class LocalFS(obnamlib.VirtualFileSystem):
                                     self.crash_counter)
 
     def reinit(self, baseurl, create=False):
-        # We fake chdir so that it doesn't mess with the caller's 
+        # We fake chdir so that it doesn't mess with the caller's
         # perception of current working directory. This also benefits
         # unit tests. To do this, we store the baseurl as the cwd.
         tracing.trace('baseurl=%s', baseurl)
@@ -101,7 +113,7 @@ class LocalFS(obnamlib.VirtualFileSystem):
         self.cwd = newcwd
 
     def lock(self, lockname, data):
-        tracing.trace('lockname=%s', lockname)
+        tracing.trace('attempting lockname=%s', lockname)
         try:
             self.write_file(lockname, data)
         except OSError, e:
@@ -109,11 +121,16 @@ class LocalFS(obnamlib.VirtualFileSystem):
                 raise obnamlib.LockFail("Lock %s already exists" % lockname)
             else:
                 raise # pragma: no cover
+        tracing.trace('got lockname=%s', lockname)
+        tracing.trace('time=%f' % time.time())
+        self.our_locks.add(lockname)
 
     def unlock(self, lockname):
         tracing.trace('lockname=%s', lockname)
-        if self.exists(lockname):
-            self.remove(lockname)
+        assert lockname in self.our_locks
+        self.remove(lockname)
+        self.our_locks.remove(lockname)
+        tracing.trace('time=%f' % time.time())
 
     def join(self, pathname):
         return os.path.join(self.cwd, pathname)
@@ -130,10 +147,10 @@ class LocalFS(obnamlib.VirtualFileSystem):
 
     def lstat(self, pathname):
         (ret, dev, ino, mode, nlink, uid, gid, rdev, size, blksize, blocks,
-         atime_sec, atime_nsec, mtime_sec, mtime_nsec, 
+         atime_sec, atime_nsec, mtime_sec, mtime_nsec,
          ctime_sec, ctime_nsec) = obnamlib._obnam.lstat(self.join(pathname))
         if ret != 0:
-            raise OSError((ret, os.strerror(ret), pathname))
+            raise OSError(ret, os.strerror(ret), pathname)
         return obnamlib.Metadata(
                     st_dev=dev,
                     st_ino=ino,
@@ -162,27 +179,38 @@ class LocalFS(obnamlib.VirtualFileSystem):
     def llistxattr(self, filename): # pragma: no cover
         ret = obnamlib._obnam.llistxattr(self.join(filename))
         if type(ret) is int:
-            raise OSError((ret, os.strerror(ret), filename))
+            raise OSError(ret, os.strerror(ret), filename)
         return [s for s in ret.split('\0') if s]
 
     def lgetxattr(self, filename, attrname): # pragma: no cover
         ret = obnamlib._obnam.lgetxattr(self.join(filename), attrname)
         if type(ret) is int:
-            raise OSError((ret, os.strerror(ret), filename))
+            raise OSError(ret, os.strerror(ret), filename)
         return ret
 
     def lsetxattr(self, filename, attrname, attrvalue): # pragma: no cover
-        ret = obnamlib._obnam.lsetxattr(self.join(filename), 
+        ret = obnamlib._obnam.lsetxattr(self.join(filename),
                                         attrname, attrvalue)
         if ret != 0:
-            raise OSError((ret, os.strerror(ret), filename))
+            raise OSError(ret, os.strerror(ret), filename)
 
     def lchown(self, pathname, uid, gid): # pragma: no cover
         tracing.trace('lchown %s %d %d', pathname, uid, gid)
         os.lchown(self.join(pathname), uid, gid)
 
-    def chmod(self, pathname, mode):
-        tracing.trace('chmod %s %o', pathname, mode)
+    # This method is excluded from test coverage because the platform
+    # either has lchmod or doesn't, and accordingly either branch of
+    # the if statement is taken, and the other branch shows up as not
+    # being tested by the unit tests.
+    def chmod_symlink(self, pathname, mode): # pragma: no cover
+        tracing.trace('chmod_symlink %s %o', pathname, mode)
+        if self.got_lchmod:
+            os.lchmod(self.join(pathname), mode)
+        else:
+            self.lstat(pathname)
+
+    def chmod_not_symlink(self, pathname, mode):
+        tracing.trace('chmod_not_symlink %s %o', pathname, mode)
         os.chmod(self.join(pathname), mode)
 
     def lutimes(self, pathname, atime_sec, atime_nsec, mtime_sec, mtime_nsec):
@@ -190,8 +218,8 @@ class LocalFS(obnamlib.VirtualFileSystem):
         assert atime_nsec is not None
         assert mtime_sec is not None
         assert mtime_nsec is not None
-        ret = obnamlib._obnam.utimensat(self.join(pathname), 
-                                        atime_sec, atime_nsec, 
+        ret = obnamlib._obnam.utimensat(self.join(pathname),
+                                        atime_sec, atime_nsec,
                                         mtime_sec, mtime_nsec)
         if ret != 0:
             raise OSError(ret, os.strerror(ret), pathname)
@@ -218,7 +246,7 @@ class LocalFS(obnamlib.VirtualFileSystem):
         tracing.trace('opened %s', pathname)
         try:
             flags = fcntl.fcntl(f.fileno(), fcntl.F_GETFL)
-            flags |= os.O_NOATIME
+            flags |= EXTRA_OPEN_FLAGS
             fcntl.fcntl(f.fileno(), fcntl.F_SETFL, flags)
         except IOError, e: # pragma: no cover
             tracing.trace('fcntl F_SETFL failed: %s', repr(e))
@@ -253,6 +281,7 @@ class LocalFS(obnamlib.VirtualFileSystem):
         self.maybe_crash()
 
     def cat(self, pathname):
+        tracing.trace('pathname=%s' % pathname)
         pathname = self.join(pathname)
         f = self.open(pathname, 'rb')
         chunks = []
@@ -266,16 +295,45 @@ class LocalFS(obnamlib.VirtualFileSystem):
         data = ''.join(chunks)
         return data
 
-    def write_file(self, pathname, contents):
+    def write_file(self, pathname, contents): # pragma: no cover
+        # This is tricky. We need to at least try to support NFS, and
+        # various filesystems that do not support hardlinks. On NFS,
+        # creating a file with O_EXCL is not guaranteed to be atomic,
+        # but adding a link with link(2) is. However, that doesn't work
+        # on VFAT, for example. So we try do both: first create a
+        # temporary file with a name guaranteed to not be a name we
+        # want to use, and then we rename it using link(2) and remove(2).
+        # If that fails, we try to create the target file with O_EXCL
+        # and rename the temporary file to that. This is still not 100%
+        # reliable: someone could be mounting VFAT across NFS, for
+        # example, but it's the best we can do. If this paragraph is
+        # wrong, tell the authors.
+
         tracing.trace('write_file %s', pathname)
         tempname = self._write_to_tempfile(pathname, contents)
         path = self.join(pathname)
+
+        # Try link(2) for creating target file.
         try:
             os.link(tempname, path)
-        except OSError, e: # pragma: no cover
+        except OSError, e:
+            pass
+        else:
+            os.remove(tempname)
+            tracing.trace('link+remove worked')
+            return
+
+        # Nope, didn't work. Now try with O_EXCL instead.
+        try:
+            fd = os.open(path, os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0666)
+            os.close(fd)
+            os.rename(tempname, path)
+        except OSError, e:
+            # Give up.
             os.remove(tempname)
             raise
-        os.remove(tempname)
+        tracing.trace('O_EXCL+rename worked')
+
         self.maybe_crash()
 
     def overwrite_file(self, pathname, contents):
@@ -284,7 +342,7 @@ class LocalFS(obnamlib.VirtualFileSystem):
         path = self.join(pathname)
         os.rename(tempname, path)
         self.maybe_crash()
-                
+
     def _write_to_tempfile(self, pathname, contents):
         path = self.join(pathname)
         dirname = os.path.dirname(path)
@@ -315,6 +373,13 @@ class LocalFS(obnamlib.VirtualFileSystem):
                 st = self.lstat(os.path.join(dirname, name))
             except OSError, e: # pragma: no cover
                 st = e
-            result.append((name, st))
-        return result
+                ino = -1
+            else:
+                ino = st.st_ino
+            result.append((ino, name, st))
+
+        # We sort things in inode order, for speed when doing namei lookups
+        # when backing up.
+        result.sort()
+        return [(name, st) for ino, name, st in result]
 
